/*

 driver.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>
 Win32 port by
  Josef Soucek <josef.soucek@ide64.org>
 Serial support by
  Borsanyi Attila <bigfoot@axelero.hu>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "driver.h"
#include <strings.h>
#include <stdlib.h>
#include "usb.h"
#include "eth.h"
#include "vice.h"
#include "rs232.h"
#include "x1541.h"
#include "pc64.h"
#include "arguments.h"

static enum e_modes mode_byname(const char *name) {
    typedef struct Mode {
        const char *name;
        enum e_modes mode;
    } Mode;
    static const Mode modes[] = {
#ifdef _X1541_H
        {"x1541", M_X1541}, {"xe1541", M_XE1541}, {"xm1541", M_XM1541}, {"xa1541", M_XA1541},
#endif
#ifdef _PC64_H
        {"pc64", M_PC64}, {"pc64s", M_PC64S},
#endif
#ifdef _RS232_H
        {"rs232", M_RS232}, {"rs232s", M_RS232S},
#endif
#ifdef _USB_H
        {"usb", M_USB},
#endif
#ifdef _VICE_H
        {"vice", M_VICE},
#endif
#ifdef _ETH_H
        {"eth", M_ETHERNET},
#endif
        {NULL, M_NONE}
    };

    if (name != NULL) {
        const Mode *i = modes;
        for (i = modes; i->name != NULL; i++) {
            if (!strcasecmp(name, i->name)) {
                return i->mode;
            }
        }
        return M_NONE;
    }
    /* select a default */
#ifdef _USB_H
    return M_USB;
#elif defined _ETH_H
    return M_ETHERNET;
#else
    return M_X1541;
#endif
}

const Driver *get_driver(Arguments *arguments) {
    const Driver *driver;
    enum e_modes mode = mode_byname(arguments->driver);

    switch (mode) {
#ifdef _X1541_H
    case M_X1541:
        driver = x1541_driver(arguments->device, arguments->lptport, arguments->hog);
        break;
    case M_XE1541:
        driver = xe1541_driver(arguments->device, arguments->lptport, arguments->hog);
        break;
    case M_XM1541:
        driver = xm1541_driver(arguments->device, arguments->lptport, arguments->hog);
        break;
    case M_XA1541:
        driver = xa1541_driver(arguments->device, arguments->lptport, arguments->hog);
        break;
#endif
#ifdef _PC64_H
    case M_PC64:
        driver = pc64_driver(arguments->device, arguments->lptport, arguments->hog);
        break;
    case M_PC64S:
        driver = pc64s_driver(arguments->device, arguments->lptport, arguments->hog);
        break;
#endif
#ifdef _RS232_H
    case M_RS232:
        driver = rs232_driver(arguments->device);
        break;
    case M_RS232S:
        driver = rs232s_driver(arguments->device);
        break;
#endif
#ifdef _USB_H
    case M_USB:
        driver = usb_driver(arguments->device);
        break;
#endif
#ifdef _VICE_H
    case M_VICE:
        driver = vice_driver(arguments->sin_addr, arguments->network);
        break;
#endif
#ifdef _ETH_H
    case M_ETHERNET:
        driver = eth_driver(arguments->sin_addr, arguments->network);
        break;
#endif
    default:
        driver = NULL;
        break;
    }
    return driver;
}
