/*

 vice.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "vice.h"
#ifdef _VICE_H
#include <string.h>
#if defined _WIN32 || defined __CYGWIN__
#include <winsock2.h>
#define SHUT_RDWR SD_BOTH
#else
#include <unistd.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#endif
#include <stdio.h>

#include "eth.h"
#include "crc8.h"
#include "log.h"
#include "driver.h"

#ifndef MSG_NOSIGNAL
#define MSG_NOSIGNAL 0
#endif

#define SERVER_IP "127.0.0.1"
#define SERVER_PORT 64245

#if defined _WIN32 || defined __CYGWIN__
#define REFUSED WSAECONNREFUSED
static SOCKET sock = INVALID_SOCKET;
static WSABUF wsabuf;
static WSAOVERLAPPED overlapped;
static int wsainited;
#define sa_family_t short
#else
#define closesocket(s) close(s)
#define INVALID_SOCKET (-1)
#define SOCKET_ERROR (-1)
#define REFUSED ECONNREFUSED
static int sock = INVALID_SOCKET;
#endif
static unsigned char ebufi[1024], ebufo[1024];
static unsigned int ebufip, ebufop, ebufil;
static const char *i_addr;
static unsigned short i_port;
static int inited;
static int driver_errno = DRIVER_EIO;

#if defined _WIN32 || defined __CYGWIN__
#define get_error() WSAGetLastError()
static const char *errormsg(int err) {
    static char msg[100];
    DWORD len = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, (DWORD)err, MAKELANGID(LANG_ENGLISH, SUBLANG_DEFAULT), (LPSTR)&msg, sizeof msg, NULL);
    if (len > 0 && msg[len - 1] == '\n') len--;
    if (len > 0 && msg[len - 1] == '\r') len--;
    if (len > 0 && msg[len - 1] == '.') len--;
    msg[len] = 0;
    return msg;
}
#else
#define get_error() (errno)
#define errormsg(err) strerror(err)
#endif

static int flush(void);

static int initialize(int lastfail) {
    static struct sockaddr_in addr;
#if defined _WIN32 || defined __CYGWIN__
    BOOL tr = 1;
#else
    int tr = 1;
#endif

    driver_errno = DRIVER_EIO;
    inited = 0;

#if defined _WIN32 || defined __CYGWIN__
    if (wsainited == 0) {
        WORD versionWanted = MAKEWORD(1, 1);
        WSADATA wsaData;
        int wsaerrno = WSAStartup(versionWanted, &wsaData);
        if (wsaerrno != 0) {
            if (lastfail != -9) {
                log_printf("Cannot start winsock: %s(%d)", errormsg(wsaerrno), wsaerrno);
            }
            return -9;
        }
        wsainited = 1;
    }
#endif

    if (addr.sin_family != AF_INET) {
        if (i_addr) {
            static struct hostent *hp;
            if (hp == NULL) {
                hp = gethostbyname(i_addr);
                if (hp == NULL) {
                    if (lastfail != -1) {
#if defined _WIN32 || defined __CYGWIN__
                        int error = get_error();
                        log_printf("Can't resolve address of %s: %s(%d)", i_addr, errormsg(error), error);
#else
                        log_printf("Can't resolve address of %s: %s(%d)", i_addr, hstrerror(h_errno), h_errno);
#endif
                    }
                    return -1;
                }
            }
            if (hp->h_addrtype != AF_INET || hp->h_addr == NULL || hp->h_length != sizeof addr.sin_addr) {
                if (lastfail != -1) {
                    log_printf("Can't resolve address of %s", i_addr);
                }
                return -1;
            }
            memcpy(&addr.sin_addr, hp->h_addr, sizeof addr.sin_addr);
        } else {
            addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
        }
        addr.sin_port = htons(i_port);
        addr.sin_family = AF_INET;
    }

    if (sock == INVALID_SOCKET) {
        sock = socket(AF_INET, SOCK_STREAM, 0);
    }
    if (sock == INVALID_SOCKET) {
        if (lastfail != -2) {
            int error = get_error();
            log_printf("Cannot open socket: %s(%d)", errormsg(error), error);
        }
        return -2;
    }
#if defined _WIN32 || defined __CYGWIN__
    if (overlapped.hEvent == WSA_INVALID_EVENT) {
        memset(&overlapped, 0, sizeof overlapped);
        overlapped.hEvent = WSACreateEvent();
        if (overlapped.hEvent == WSA_INVALID_EVENT) {
            if (lastfail != -3) {
                int error = get_error();
                log_printf("WSACreateEvent failed: %s(%d)", errormsg(error), error);
            }
            return -3;
        }
    }
#endif

    if (connect(sock, (struct sockaddr *)&addr, sizeof addr) == SOCKET_ERROR) {
        if (lastfail != -4) {
            int error = get_error();
            if (error == REFUSED) {
                log_printf("Waiting for VICE to show up");
            } else {
                log_printf("Cannot connect to %s:%u: %s(%d)", i_addr != NULL ? i_addr : "127.0.0.1", i_port, errormsg(error), error);
            }
        }
        closesocket(sock);
        sock = INVALID_SOCKET;
        return -4;
    }
    if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char *)&tr, sizeof tr) == SOCKET_ERROR) {
        int error = get_error();
        log_printf("Setting socket options failed: %s(%d)", errormsg(error), error);
    }

#if defined _WIN32 || defined __CYGWIN__
    memset(&wsabuf, 0, sizeof wsabuf);
    wsabuf.buf = (char *)ebufi;
    wsabuf.len = sizeof ebufi;
#endif
    log_printf("Connected to %s:%u", i_addr != NULL ? i_addr : "127.0.0.1", i_port);
    inited = 1;
    return 0;
}

static int refill(void) {
    if (driver_errno == DRIVER_SUCCESS) {
#if defined _WIN32 || defined __CYGWIN__
        int ret;
        DWORD n, f = 0;
        ebufip = ebufil = 0;
        WSAResetEvent(overlapped.hEvent);
        ret = WSARecv(sock, &wsabuf, 1, &n, &f, &overlapped, NULL);
        if (ret == SOCKET_ERROR) { 
            if (WSAGetLastError() != WSA_IO_PENDING) return DRIVER_EIO;
            ret = WSAWaitForMultipleEvents(1, &overlapped.hEvent, TRUE, INFINITE, TRUE);
            if (ret == WSA_WAIT_FAILED) return DRIVER_EIO;
            ret = WSAGetOverlappedResult(sock, &overlapped, &n, FALSE, &f);
            if (ret == FALSE) return DRIVER_EIO;
        } else if (ret != 0) return DRIVER_EIO;
        if (n == 0) return DRIVER_EIO;
#else
        int n = recv(sock, (char *)ebufi, sizeof ebufi, 0);
        ebufip = ebufil = 0;
        if (n == SOCKET_ERROR || n <= 0) return DRIVER_EIO;
#endif
        ebufil = (unsigned int)n;
    }
    return driver_errno;
}

static int getb(int use_timeout) {
    unsigned char a;
    (void)use_timeout;
    if (ebufip >= sizeof ebufi || ebufip >= ebufil) {
        driver_errno = refill();
        if (driver_errno != DRIVER_SUCCESS) return EOF;
    }
    a = ebufi[ebufip++];
    crc_add_byte(a);
    return a;
}

static void sendb(unsigned char a) {
    if (ebufop >= sizeof ebufo) {
        driver_errno = flush();
        if (driver_errno != DRIVER_SUCCESS) return;
    }
    crc_add_byte(a);
    ebufo[ebufop++] = a;
}

static void getbytes(unsigned char data[], unsigned int bytes) {
    while (ebufip + bytes > ebufil) {
        ebufil -= ebufip;
        memcpy(data, ebufi + ebufip, ebufil);
        crc_add_block(data, ebufil);
        data += ebufil; bytes -= ebufil;
        driver_errno = refill();
        if (driver_errno != DRIVER_SUCCESS) return;
    }
    memcpy(data, ebufi + ebufip, bytes);
    crc_add_block(data, bytes);
    ebufip += bytes;
}

static void sendbytes(const unsigned char data[], unsigned int bytes) {
    crc_add_block(data, bytes);
    while (ebufop + bytes > sizeof ebufo) {
        memcpy(ebufo + ebufop, data, sizeof(ebufo) - ebufop);
        data += sizeof(ebufo) - ebufop;
        bytes -= sizeof(ebufo) - ebufop;
        ebufop = sizeof ebufo;
        driver_errno = flush();
        if (driver_errno != DRIVER_SUCCESS) return;
    }
    memcpy(ebufo + ebufop, data, bytes);
    ebufop += bytes;
}

static void eshutdown(void) {
    if (sock != INVALID_SOCKET) {
        if (inited) {
            shutdown(sock, SHUT_RDWR);
        }
        closesocket(sock);
        sock = INVALID_SOCKET;
    }
#if defined _WIN32 || defined __CYGWIN__
    if (overlapped.hEvent != WSA_INVALID_EVENT) {
        WSACloseEvent(overlapped.hEvent);
        overlapped.hEvent = WSA_INVALID_EVENT;
    }
    if (wsainited == 1) {
        WSACleanup();
        wsainited = 0;
    }
#endif
    inited = 0;
    return;
}

static int flush(void) {
    if (ebufop && driver_errno == DRIVER_SUCCESS) {
        if (send(sock, (char *)ebufo, ebufop, MSG_NOSIGNAL) == SOCKET_ERROR) {
            int error = get_error();
            log_printf("Couldn't send: %s(%d)", errormsg(error), error);
            ebufop = 0;
            return DRIVER_EIO;
        }
        ebufop = 0;
    }
    return driver_errno;
}

static int done(void) {
    return driver_errno;
}

static void turn(void) {
}

static int clean(void) {
    ebufip = 0;
    ebufop = 0;
    ebufil = 0;
    return 0;
}

static int waitb(unsigned char ec) {
    int i;
    (void)ec;
    if (!inited) return DRIVER_ENODEV;
    driver_errno = DRIVER_SUCCESS;
    i = getb(1);
    return (i != EOF) ? i : driver_errno;
}

static const Driver driver = {
    .mode         = M_VICE,
    .name         = "VICE",
    .initialize   = initialize,
    .getb         = getb,
    .sendb        = sendb,
    .getbytes     = getbytes,
    .sendbytes    = sendbytes,
    .shutdown     = eshutdown,
    .flush        = flush,
    .done         = done,
    .turn         = turn,
    .wait         = waitb,
    .clean        = clean,
};

const Driver *vice_driver(const char *addr, int port) {
    i_addr = addr;
    i_port = (port > 0 && port <= 65535) ? (unsigned short)port : SERVER_PORT;
    log_printf("Using %s driver connecting to %s:%u", driver.name, i_addr ? i_addr : "127.0.0.1", i_port);
    return &driver;
}
#endif
